% dfAlignSIFT - Affine image alignment using DF SIFT
%   [fit distance] = dfAlignSIFT(img, tmplt, p_init, NUM_ITERATIONS,downsampleScale,...
%    sigmaXYChoices,direction)
%   Align the template image tmplt to an example image img using an
%   affine warp initialized using p_init.
%
%   This assumes grayscale images and rectangular templates.
%
%    Inputs:
%        img: The image being aligned to tmplt
%        tmplt: The region that img is being aligned to
%        p_init: The initial warp to apply to img
%        NUM_ITERATIONS: Maximum number of iterations
%        downsampleScale: How much to downsample the distribution
%        fields in spatial dimensions
%        sigmaXYChoices: Set of xy-sigmas to search over when
%        constructing the DF SIFT
%        direction: whether to perform forward (0),
%        inverse (1), or inverse-forward (2);
%
%    Outputs:
%        fit: Struct containing the intermediate and final transformations
%        xySigma: Final xySigma that minimized the L2 distance between blurred template
%        DF SIFT and warped input image DF SIFT
%        distance: Final minimal L2 distance between blurred template DF
%        SIFT and warped input image DF SIFT
%
%
% Requires vlfeat toolbox (located in same directory)

function [fit xySigma distance] = dfAlignSIFT(img, tmplt, p_init, NUM_ITERATIONS,downsampleScale,...
    sigmaXYChoices,direction)
if(~isdeployed)
    addpath('vlfeat/toolbox');
    vl_setup;
end
height = size(tmplt,1); width = size(tmplt,2);


siftBinSize = 4;
siftPatchSize = siftBinSize*3;

featureDim = 128;
numBins = featureDim;
[f d] = vl_dsift(single(tmplt),'size',siftBinSize);
d = reshape(d,featureDim,height-siftPatchSize,width-siftPatchSize);
d = shiftdim(d,1);
d= double(d);

siftHeight = height-siftPatchSize;
siftWidth = width - siftPatchSize;

if(~exist('direction','var')) direction = 2; end


fit = struct([]);
T = [p_init' [0;0;1]];

heightNew = floor(height/downsampleScale);
widthNew = floor(width/downsampleScale);

siftHeightNew = floor(siftHeight/downsampleScale);
siftWidthNew = floor(siftWidth/downsampleScale);

%Convert p_init to downsampled coord system
pntsOrig = [1 1 1;1 height 1;width 1 1;width height 1];
pntsNew = pntsOrig* T;
pntsNew = pntsNew./repmat(pntsNew(:,3),[1 3]);
downSamplePntsOrig = pntsOrig/downsampleScale;
downSamplePntsOrig(:,3)=1;
downSampledPntsNew = [ pntsNew(:,1:2)/downsampleScale ones(4,1)];
T = pinv(downSamplePntsOrig)*downSampledPntsNew;

%precompute distribution fields for template
templateDistributionFields = zeros(siftHeightNew,siftWidthNew,numBins,size(sigmaXYChoices,2),1);
templateDistributionFields2 = zeros(siftHeight,siftWidth,numBins,size(sigmaXYChoices,2),1);

for i=1:size(sigmaXYChoices,2)
    xysigma = sigmaXYChoices(i);
    xysize = max(4*sigmaXYChoices(i)+1,3);
    filter1=reshape(makeGaussianKernel(xysize,xysigma),[xysize 1]);
    filter2=reshape(makeGaussianKernel(xysize,xysigma),[1 xysize]);
    tempDF = convn(convn(d,filter1,'same'),filter2,'same');
    templateDistributionFields2(:,:,:,i,1) = tempDF;
    templateDistributionFields(:,:,:,i,1) = tempDF(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
end


degFreedom = 6;
dWdP = zeros(2,degFreedom,siftHeightNew,siftWidthNew);

dWdPAll = zeros(2*numBins*siftHeightNew*siftWidthNew,6);
for j=1:siftWidthNew
    for i=1:siftHeightNew
        dWdP(:,:,i,j)=[j 0 i 0 1 0;
            0 j 0 i 0 1];
        dWdPAll((i-1)*numBins*2+siftHeightNew*numBins*2*(j-1)+1:...
            (i-1)*numBins*2+siftHeightNew*numBins*2*(j-1)+numBins*2,:) = ...
            repmat([j 0 i 0 1 0;0 j 0 i 0 1],[numBins 1]);
    end
end

dWdPReshaped = zeros(siftHeightNew*siftWidthNew*numBins,2,6);
for i=1:6
    dWdPReshaped(:,:,i) = reshape(dWdPAll(:,i),2,[])';
end

%Precompute Steepest Descent Images and Hessian
featureSpaceDim = numBins;

steepestDescentImages = zeros(siftHeightNew*siftWidthNew*featureSpaceDim,degFreedom,...
    size(sigmaXYChoices,2),1);
Hessians = zeros(degFreedom,degFreedom,size(sigmaXYChoices,2),1);
InverseHessians = zeros(degFreedom,degFreedom,size(sigmaXYChoices,2),1);

for m=1:size(sigmaXYChoices,2)
    [Ex,Ey] = gradient(templateDistributionFields(:,:,:,m,1));
    
    ExReshaped = shiftdim(Ex,2); EyReshaped = shiftdim(Ey,2);
    ExEy = [ExReshaped(:) EyReshaped(:)];
    for i=1:6
        steepestDescentImages(:,i,m,1) = sum(ExEy.*dWdPReshaped(:,:,i),2);
    end
    
    Hessians(:,:,m,1) = steepestDescentImages(:,:,m,1)'*steepestDescentImages(:,:,m,1);
end

%Precompute Inverses of Hessians
for m=1:size(sigmaXYChoices,2)
    InverseHessians(:,:,m,1) = Hessians(:,:,m,1)^(-1);
end


additiveNoise = 0; gain = 0;
iteration =1;
while(true)
    %Compute warp in non-downsampled image
    downSampledPntsOrig = [1 1 1;1 heightNew 1;widthNew 1 1;widthNew heightNew 1];
    downSampledPntsNew = downSampledPntsOrig* T;
    downSampledPntsNew = downSampledPntsNew./repmat(downSampledPntsNew(:,3),[1 3]);
    pntsOrig = [downsampleScale* downSampledPntsOrig(:,1:2) ones(4,1)];
    pntsNew = [downsampleScale* downSampledPntsNew(:,1:2) ones(4,1)];
    T_temp = pinv(pntsOrig)*pntsNew;
    fit(iteration).warp_p = T_temp'; %save transpose to agree with Lucas-Kanade code convention
    
    currentWarpedImage = (imtransform2(img,T_temp)*(gain+1))+additiveNoise;
    currentWarpedImage = currentWarpedImage(1:height,1:width);
    
    currentWarpedImage(currentWarpedImage>255)=255;
    currentWarpedImage(currentWarpedImage<0)=0;
    
    
    [f d] = vl_dsift(single(currentWarpedImage),'size',siftBinSize);
    d = double(d);
    dWarped = reshape(d,featureDim,height-siftPatchSize,width-siftPatchSize);
    dWarped = shiftdim(dWarped,1);
    
    [bestXYChoice] = findBestSigma(templateDistributionFields2,dWarped,sigmaXYChoices);
    fit(iteration).xySigma = sigmaXYChoices(bestXYChoice);
    fit(iteration).fSigma = 0; %keep field for consistency with dfAlign code but always assign 0
    %fprintf('\niteration #%d bestXYChoice = %.1f\n',iteration,sigmaXYChoices(bestXYChoice))
    
    xysigma = sigmaXYChoices(bestXYChoice);
    xysize = max(4*sigmaXYChoices(bestXYChoice)+1,3);
    filter1=reshape(makeGaussianKernel(xysize,xysigma),[xysize 1]);
    filter2=reshape(makeGaussianKernel(xysize,xysigma),[1 xysize]);
    dWarped = convn(convn(dWarped,filter1,'same'),filter2,'same');
    %fprintf('\niteration %d distance = %.4f\n',iteration,sum(sum(sum((templateDistributionFields2(:,:,:,1,1)-dWarped).^2))));
    
    warpedDF = dWarped;
    warpedDF = warpedDF(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
    
    wDF = shiftdim(warpedDF,2);
    tDF = shiftdim(templateDistributionFields(:,:,:,bestXYChoice,1),2);
    
    if(direction == 1 || direction ==2)
        deltaPInverse = InverseHessians(:,:,bestXYChoice,1)*...
            sum(steepestDescentImages(:,:,bestXYChoice,1) .*...
            repmat(wDF(:)-tDF(:),[1 degFreedom]),1)';
        deltaPTemp =reshape([deltaPInverse(1:6)],[2,3])';
        incrementalWarpInverse = [1 0 0;0 1 0;0 0 1] + [deltaPTemp [0;0;0]];
    end
    
    if(direction == 0 || direction ==2)
        [Ex2,Ey2] = gradient(warpedDF);
        steepestDescentImages2 = zeros(numBins*siftHeightNew*siftWidthNew,degFreedom);
        ExReshaped2 = shiftdim(Ex2,2); EyReshaped2 = shiftdim(Ey2,2);
        ExEy = [ExReshaped2(:) EyReshaped2(:)];
        for i=1:6
            steepestDescentImages2(:,i) = sum(ExEy.*dWdPReshaped(:,:,i),2);
        end
        
        Hessian2= steepestDescentImages2'*steepestDescentImages2;
        %Compute delta p for forward
        deltaPForward = Hessian2^(-1)*sum(steepestDescentImages2.*repmat(tDF(:)-wDF(:),[1 degFreedom]),1)';
    end
    
    
    if(direction == 0 || direction ==2)
        deltaP2Temp =reshape([deltaPForward(1:6)],[2,3])';
        incrementalWarpForward = [1 0 0;0 1 0;0 0 1] + [deltaP2Temp [0;0;0]];
    end
    
    if(direction ==2)
        incrementalWarp = real(expm((logm(incrementalWarpInverse^(-1))+ ...
            logm(incrementalWarpForward^(1)))/2));
    elseif(direction ==1)
        incrementalWarp = incrementalWarpInverse^(-1);
    else
        incrementalWarp = incrementalWarpForward;
    end
    update = incrementalWarp;
    T = update*T;

    terminationEPS = .003;
    
    if(iteration==NUM_ITERATIONS || sqrt(sum(sum((update-eye(3)).^2)))<terminationEPS)
        break;
    end
    iteration = iteration +1;
end %end while(true)

%Convert T to original coords
downSampledPntsOrig = [1 1 1;1 heightNew 1;widthNew 1 1;widthNew heightNew 1];
downSampledPntsNew = downSampledPntsOrig* T;
downSampledPntsNew = downSampledPntsNew./repmat(downSampledPntsNew(:,3),[1 3]);
pntsOrig = [downsampleScale* downSampledPntsOrig(:,1:2) ones(4,1)];
pntsNew = [downsampleScale* downSampledPntsNew(:,1:2) ones(4,1)];
T = pinv(pntsOrig)*pntsNew;
fit(iteration).warp_p =  T'; %save transpose to agree with Lucas-Kanade convention
fit = fit(1:iteration);

currentWarpedImage = (imtransform2(img,T)*(gain+1))+additiveNoise;
currentWarpedImage = currentWarpedImage(1:height,1:width);

currentWarpedImage(currentWarpedImage>255)=255;
currentWarpedImage(currentWarpedImage<0)=0;


[f d] = vl_dsift(single(currentWarpedImage),'size',siftBinSize);
d = double(d);
dWarped = reshape(d,featureDim,height-siftPatchSize,width-siftPatchSize);
dWarped = shiftdim(dWarped,1);

[bestXYChoice] = findBestSigma(templateDistributionFields2,dWarped,sigmaXYChoices);
    
xySigma = sigmaXYChoices(bestXYChoice);
xysize = max(4*sigmaXYChoices(bestXYChoice)+1,3);
filter1=reshape(makeGaussianKernel(xysize,xySigma),[xysize 1]);
filter2=reshape(makeGaussianKernel(xysize,xySigma),[1 xysize]);
dWarped = convn(convn(dWarped,filter1,'same'),filter2,'same');
distance = sum(sum(sum((templateDistributionFields2(:,:,:,1,1)-dWarped).^2)));

function [bestXYChoice, dist] = findBestSigma(templateDFs,currentDF,sigmaXYChoices)

distances = zeros(numel(sigmaXYChoices),1);

for m=1:size(sigmaXYChoices,2)
    dists = (templateDFs(:,:,:,m,1)-currentDF).^2;
    dist = sum(dists(:));
    distances(m,1)=dist;
end

[val bestXYChoice] = min(distances);

dist = val;
